@@ -17,10 +17,10 @@
 #include <spa/pod/filter.h>
 #include <spa/pod/dynamic.h>
 #include <spa/debug/types.h>
-#include <spa/debug/dict.h>
 
 #define PW_ENABLE_DEPRECATED
 
+#include <pipewire/cleanup.h>
 #include "pipewire/pipewire.h"
 #include "pipewire/stream.h"
 #include "pipewire/private.h"
@@ -83,13 +83,13 @@ struct stream {
 	const char *path;
 
 	struct pw_context *context;
-	struct spa_hook context_listener;
+
+	struct pw_loop *main_loop;
+	struct pw_loop *data_loop;
 
 	enum spa_direction direction;
 	enum pw_stream_flags flags;
 
-	struct pw_impl_node *node;
-
 	struct spa_node impl_node;
 	struct spa_node_methods node_methods;
 	struct spa_hook_list hooks;
@@ -113,7 +113,8 @@ struct stream {
 #define PORT_Format	3
 #define PORT_Buffers	4
 #define PORT_Latency	5
-#define N_PORT_PARAMS	6
+#define PORT_Tag	6
+#define N_PORT_PARAMS	7
 	struct spa_param_info port_params[N_PORT_PARAMS];
 
 	struct spa_list param_list;
@@ -156,7 +157,9 @@ struct stream {
 	unsigned int driving:1;
 	unsigned int using_trigger:1;
 	unsigned int trigger:1;
+	unsigned int early_process:1;
 	int in_set_param;
+	int in_emit_param_changed;
 };
 
 static int get_param_index(uint32_t id)
@@ -190,15 +193,17 @@ static int get_port_param_index(uint32_t id)
 		return PORT_Buffers;
 	case SPA_PARAM_Latency:
 		return PORT_Latency;
+	case SPA_PARAM_Tag:
+		return PORT_Tag;
 	default:
 		return -1;
 	}
 }
 
-static void fix_datatype(const struct spa_pod *param)
+static void fix_datatype(struct spa_pod *param)
 {
 	const struct spa_pod_prop *pod_param;
-	const struct spa_pod *vals;
+	struct spa_pod *vals;
 	uint32_t dataType, n_vals, choice;
 
 	pod_param = spa_pod_find_prop(param, NULL, SPA_PARAM_BUFFERS_dataType);
@@ -238,17 +243,17 @@ static struct param *add_param(struct stream *impl,
 	if (p == NULL)
 		return NULL;
 
-	if (id == SPA_PARAM_Buffers &&
-	    SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_MAP_BUFFERS) &&
-	    impl->direction == SPA_DIRECTION_INPUT)
-		fix_datatype(param);
-
 	p->id = id;
 	p->flags = flags;
 	p->param = SPA_PTROFF(p, sizeof(struct param), struct spa_pod);
 	memcpy(p->param, param, SPA_POD_SIZE(param));
 	SPA_POD_OBJECT_ID(p->param) = id;
 
+	if (id == SPA_PARAM_Buffers &&
+	    SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_MAP_BUFFERS) &&
+	    impl->direction == SPA_DIRECTION_INPUT)
+		fix_datatype(p->param);
+
 	spa_list_append(&impl->param_list, &p->link);
 
 	if ((idx = get_param_index(id)) != -1) {
@@ -267,14 +272,42 @@ static struct param *add_param(struct stream *impl,
 static void clear_params(struct stream *impl, uint32_t id)
 {
 	struct param *p, *t;
+	bool found = false;
+	int i, idx;
 
 	spa_list_for_each_safe(p, t, &impl->param_list, link) {
 		if (id == SPA_ID_INVALID ||
 		    (p->id == id && !(p->flags & PARAM_FLAG_LOCKED))) {
+			found = true;
 			spa_list_remove(&p->link);
 			free(p);
 		}
 	}
+	if (found) {
+		if (id == SPA_ID_INVALID) {
+			impl->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
+			for (i = 0; i < N_NODE_PARAMS; i++) {
+				impl->params[i].flags &= ~SPA_PARAM_INFO_READ;
+				impl->params[i].user++;
+			}
+			impl->port_info.change_mask |= SPA_PORT_CHANGE_MASK_PARAMS;
+			for (i = 0; i < N_PORT_PARAMS; i++) {
+				impl->port_params[i].flags &= ~SPA_PARAM_INFO_READ;
+				impl->port_params[i].user++;
+			}
+		} else {
+			if ((idx = get_param_index(id)) != -1) {
+				impl->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
+				impl->params[idx].flags &= ~SPA_PARAM_INFO_READ;
+				impl->params[idx].user++;
+			}
+			if ((idx = get_port_param_index(id)) != -1) {
+				impl->port_info.change_mask |= SPA_PORT_CHANGE_MASK_PARAMS;
+				impl->port_params[idx].flags &= ~SPA_PARAM_INFO_READ;
+				impl->port_params[idx].user++;
+			}
+		}
+	}
 }
 
 static int update_params(struct stream *impl, uint32_t id,
@@ -351,26 +384,28 @@ static inline void clear_queue(struct stream *stream, struct queue *queue)
 	queue->incount = queue->outcount;
 }
 
-static bool stream_set_state(struct pw_stream *stream, enum pw_stream_state state, const char *error)
+static bool stream_set_state(struct pw_stream *stream, enum pw_stream_state state,
+		int res, const char *error)
 {
 	enum pw_stream_state old = stream->state;
-	bool res = old != state;
+	bool changed = old != state;
 
-	if (res) {
+	if (changed) {
 		free(stream->error);
 		stream->error = error ? strdup(error) : NULL;
+		stream->error_res = res;
 
-		pw_log_debug("%p: update state from %s -> %s (%s)", stream,
+		pw_log_debug("%p: update state from %s -> %s (%d) %s", stream,
 			     pw_stream_state_as_string(old),
-			     pw_stream_state_as_string(state), stream->error);
+			     pw_stream_state_as_string(state), res, stream->error);
 
 		if (state == PW_STREAM_STATE_ERROR)
-			pw_log_error("%p: error %s", stream, error);
+			pw_log_error("%p: error (%d) %s", stream, res, error);
 
 		stream->state = state;
 		pw_stream_emit_state_changed(stream, old, state, error);
 	}
-	return res;
+	return changed;
 }
 
 static struct buffer *get_buffer(struct pw_stream *stream, uint32_t id)
@@ -403,7 +438,7 @@ static inline uint32_t update_requested(struct stream *impl)
 		buffer->this.requested = impl->quantum;
 		res = 1;
 	}
-	pw_log_trace_fp("%p: update buffer:%u size:%"PRIu64, impl, id, buffer->this.requested);
+	pw_log_trace_fp("%p: update buffer:%u req:%"PRIu64, impl, id, buffer->this.requested);
 	return res;
 }
 
@@ -414,20 +449,24 @@ do_call_process(struct spa_loop *loop,
 	struct stream *impl = user_data;
 	struct pw_stream *stream = &impl->this;
 	pw_log_trace_fp("%p: do process", stream);
-	pw_stream_emit_process(stream);
+	if (!impl->disconnecting)
+		pw_stream_emit_process(stream);
 	return 0;
 }
 
 static inline void call_process(struct stream *impl)
 {
 	pw_log_trace_fp("%p: call process rt:%u", impl, impl->process_rt);
-	if (impl->direction == SPA_DIRECTION_OUTPUT && update_requested(impl) <= 0)
+	if (impl->n_buffers == 0 ||
+	    (impl->direction == SPA_DIRECTION_OUTPUT && update_requested(impl) <= 0))
 		return;
-	if (impl->process_rt)
-		spa_callbacks_call(&impl->rt_callbacks, struct pw_stream_events, process, 0);
-	else
-		pw_loop_invoke(impl->context->main_loop,
+	if (impl->process_rt) {
+		if (impl->rt_callbacks.funcs)
+			spa_callbacks_call_fast(&impl->rt_callbacks, struct pw_stream_events, process, 0);
+	} else {
+		pw_loop_invoke(impl->main_loop,
 			do_call_process, 1, NULL, 0, false, impl);
+	}
 }
 
 static int
@@ -443,7 +482,8 @@ do_call_drained(struct spa_loop *loop,
 
 static void call_drained(struct stream *impl)
 {
-	pw_loop_invoke(impl->context->main_loop,
+	pw_log_info("%p: drained", impl);
+	pw_loop_invoke(impl->main_loop,
 		do_call_drained, 1, NULL, 0, false, impl);
 }
 
@@ -460,7 +500,7 @@ do_call_trigger_done(struct spa_loop *loop,
 
 static void call_trigger_done(struct stream *impl)
 {
-	pw_loop_invoke(impl->context->main_loop,
+	pw_loop_invoke(impl->main_loop,
 		do_call_trigger_done, 1, NULL, 0, false, impl);
 }
 
@@ -494,7 +534,7 @@ static int impl_set_io(void *object, uint32_t id, void *data, size_t size)
 		else
 			impl->position = NULL;
 
-		pw_loop_invoke(impl->context->data_loop,
+		pw_loop_invoke(impl->data_loop,
 				do_set_position, 1, NULL, 0, true, impl);
 		break;
 	default:
@@ -558,16 +598,24 @@ static int impl_enum_params(void *object, int seq, uint32_t id, uint32_t start,
 	return enum_params(object, false, seq, id, start, num, filter);
 }
 
+static inline void emit_param_changed(struct stream *impl,
+		uint32_t id, const struct spa_pod *param)
+{
+	struct pw_stream *stream = &impl->this;
+	if (impl->in_emit_param_changed++ == 0)
+		pw_stream_emit_param_changed(stream, id, param);
+	impl->in_emit_param_changed--;
+}
+
 static int impl_set_param(void *object, uint32_t id, uint32_t flags, const struct spa_pod *param)
 {
 	struct stream *impl = object;
-	struct pw_stream *stream = &impl->this;
 
 	if (id != SPA_PARAM_Props)
 		return -ENOTSUP;
 
 	if (impl->in_set_param == 0)
-		pw_stream_emit_param_changed(stream, id, param);
+		emit_param_changed(impl, id, param);
 
 	return 0;
 }
@@ -576,7 +624,7 @@ static inline void copy_position(struct stream *impl, int64_t queued)
 {
 	struct spa_io_position *p = impl->rt.position;
 
-	SEQ_WRITE(impl->seq);
+	SPA_SEQ_WRITE(impl->seq);
 	if (SPA_LIKELY(p != NULL)) {
 		impl->time.now = p->clock.nsec;
 		impl->time.rate = p->clock.rate;
@@ -591,7 +639,7 @@ static inline void copy_position(struct stream *impl, int64_t queued)
 	}
 	if (SPA_LIKELY(impl->rate_match != NULL))
 		impl->rate_queued = impl->rate_match->delay;
-	SEQ_WRITE(impl->seq);
+	SPA_SEQ_WRITE(impl->seq);
 }
 
 static int impl_send_command(void *object, const struct spa_command *command)
@@ -607,12 +655,12 @@ static int impl_send_command(void *object, const struct spa_command *command)
 	case SPA_NODE_COMMAND_Suspend:
 	case SPA_NODE_COMMAND_Flush:
 	case SPA_NODE_COMMAND_Pause:
-		pw_loop_invoke(impl->context->main_loop,
+		pw_loop_invoke(impl->main_loop,
 			NULL, 0, NULL, 0, false, impl);
 		if (stream->state == PW_STREAM_STATE_STREAMING) {
 
 			pw_log_debug("%p: pause", stream);
-			stream_set_state(stream, PW_STREAM_STATE_PAUSED, NULL);
+			stream_set_state(stream, PW_STREAM_STATE_PAUSED, 0, NULL);
 		}
 		break;
 	case SPA_NODE_COMMAND_Start:
@@ -628,7 +676,7 @@ static int impl_send_command(void *object, const struct spa_command *command)
 				call_process(impl);
 			}
 
-			stream_set_state(stream, PW_STREAM_STATE_STREAMING, NULL);
+			stream_set_state(stream, PW_STREAM_STATE_STREAMING, 0, NULL);
 		}
 		break;
 	default:
@@ -814,7 +862,7 @@ static void clear_buffers(struct pw_stream *stream)
 
 		while ((b = queue_pop(impl, &impl->dequeued))) {
 			if (b->busy)
-				ATOMIC_DEC(b->busy->count);
+				SPA_ATOMIC_DEC(b->busy->count);
 		}
 	} else
 		clear_queue(impl, &impl->dequeued);
@@ -880,10 +928,10 @@ static int impl_port_set_param(void *object,
 		break;
 	}
 
-	pw_stream_emit_param_changed(stream, id, param);
+	emit_param_changed(impl, id, param);
 
 	if (stream->state == PW_STREAM_STATE_ERROR)
-		return -EIO;
+		return stream->error_res;
 
 	emit_node_info(impl, false);
 	emit_port_info(impl, false);
@@ -993,12 +1041,15 @@ static int impl_node_process_input(void *object)
 		/* push new buffer */
 		pw_log_trace_fp("%p: push %d %p", stream, b->id, io);
 		if (queue_push(impl, &impl->dequeued, b) == 0) {
-			copy_position(impl, impl->dequeued.incount);
 			if (b->busy)
-				ATOMIC_INC(b->busy->count);
-			call_process(impl);
+				SPA_ATOMIC_INC(b->busy->count);
 		}
 	}
+	if (!queue_is_empty(impl, &impl->dequeued)) {
+		copy_position(impl, impl->dequeued.incount);
+		call_process(impl);
+	}
+
 	if (io->status != SPA_STATUS_NEED_DATA || io->buffer_id == SPA_ID_INVALID) {
 		/* pop buffer to recycle */
 		if ((b = queue_pop(impl, &impl->queued))) {
@@ -1053,7 +1104,7 @@ again:
 			 * rate matching node (audioconvert) has been scheduled to
 			 * update the values. */
 			ask_more = !impl->process_rt && impl->rate_match == NULL &&
-				queue_is_empty(impl, &impl->queued) &&
+				(impl->early_process || queue_is_empty(impl, &impl->queued)) &&
 				!queue_is_empty(impl, &impl->dequeued);
 			pw_log_trace_fp("%p: pop %d %p ask_more:%u %p", stream, b->id, io,
 					ask_more, impl->rate_match);
@@ -1071,7 +1122,7 @@ again:
 		}
 	} else {
 		ask_more = !impl->process_rt &&
-			queue_is_empty(impl, &impl->queued) &&
+			(impl->early_process || queue_is_empty(impl, &impl->queued)) &&
 			!queue_is_empty(impl, &impl->dequeued);
 	}
 
@@ -1119,7 +1170,7 @@ static void proxy_removed(void *_data)
 	pw_log_debug("%p: removed", stream);
 	spa_hook_remove(&stream->proxy_listener);
 	stream->node_id = SPA_ID_INVALID;
-	stream_set_state(stream, PW_STREAM_STATE_UNCONNECTED, NULL);
+	stream_set_state(stream, PW_STREAM_STATE_UNCONNECTED, 0, NULL);
 }
 
 static void proxy_destroy(void *_data)
@@ -1145,7 +1196,7 @@ static void proxy_bound_props(void *data, uint32_t global_id, const struct spa_d
 	stream->node_id = global_id;
 	if (props)
 		pw_properties_update(stream->properties, props);
-	stream_set_state(stream, PW_STREAM_STATE_PAUSED, NULL);
+	stream_set_state(stream, PW_STREAM_STATE_PAUSED, 0, NULL);
 }
 
 static const struct pw_proxy_events proxy_events = {
@@ -1341,13 +1392,14 @@ static void node_event_destroy(void *data)
 	struct pw_stream *stream = data;
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	spa_hook_remove(&stream->node_listener);
-	impl->node = NULL;
+	pw_impl_node_remove_rt_listener(stream->node, &stream->node_rt_listener);
+	stream->node = NULL;
+	impl->data_loop = NULL;
 }
 
 static void node_event_info(void *data, const struct pw_node_info *info)
 {
 	struct pw_stream *stream = data;
-	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	uint32_t i;
 
 	if (info->change_mask & PW_NODE_CHANGE_MASK_PARAMS) {
@@ -1355,7 +1407,7 @@ static void node_event_info(void *data, const struct pw_node_info *info)
 			switch (info->params[i].id) {
 			case SPA_PARAM_PropInfo:
 			case SPA_PARAM_Props:
-				pw_impl_node_for_each_param(impl->node,
+				pw_impl_node_for_each_param(stream->node,
 						0, info->params[i].id,
 						0, UINT32_MAX,
 						NULL,
@@ -1383,7 +1435,7 @@ static void on_core_error(void *data, uint32_t id, int seq, int res, const char
 			id, seq, res, spa_strerror(res), message);
 
 	if (id == PW_ID_CORE && res == -EPIPE) {
-		stream_set_state(stream, PW_STREAM_STATE_UNCONNECTED, message);
+		stream_set_state(stream, PW_STREAM_STATE_UNCONNECTED, res, message);
 	}
 }
 
@@ -1392,11 +1444,9 @@ static const struct pw_core_events core_events = {
 	.error = on_core_error,
 };
 
-static void context_drained(void *data, struct pw_impl_node *node)
+static void node_drained(void *data)
 {
 	struct stream *impl = data;
-	if (impl->node != node)
-		return;
 	if (impl->draining && impl->drained) {
 		impl->draining = false;
 		if (impl->io != NULL)
@@ -1405,9 +1455,9 @@ static void context_drained(void *data, struct pw_impl_node *node)
 	}
 }
 
-static const struct pw_context_driver_events context_events = {
-	PW_VERSION_CONTEXT_DRIVER_EVENTS,
-	.drained = context_drained,
+static const struct pw_impl_node_rt_events node_rt_events = {
+	PW_VERSION_IMPL_NODE_RT_EVENTS,
+	.drained = node_drained,
 };
 
 struct match {
@@ -1433,7 +1483,6 @@ stream_new(struct pw_context *context, const char *name,
 	struct stream *impl;
 	struct pw_stream *this;
 	const char *str;
-	struct match match;
 	int res;
 
 	ensure_loop(context->main_loop, return NULL);
@@ -1448,6 +1497,7 @@ stream_new(struct pw_context *context, const char *name,
 		res = -errno;
 		goto error_properties;
 	}
+	impl->main_loop = pw_context_get_main_loop(context);
 
 	this = &impl->this;
 	pw_log_debug("%p: new \"%s\"", impl, name);
@@ -1464,28 +1514,6 @@ stream_new(struct pw_context *context, const char *name,
 	spa_hook_list_init(&impl->hooks);
 	this->properties = props;
 
-	pw_context_conf_update_props(context, "stream.properties", props);
-
-	match = MATCH_INIT(this);
-	pw_context_conf_section_match_rules(context, "stream.rules",
-			&this->properties->dict, execute_match, &match);
-
-	if ((str = getenv("PIPEWIRE_PROPS")) != NULL)
-		pw_properties_update_string(props, str, strlen(str));
-	if ((str = getenv("PIPEWIRE_QUANTUM")) != NULL) {
-		struct spa_fraction q;
-		if (sscanf(str, "%u/%u", &q.num, &q.denom) == 2 && q.denom != 0) {
-			pw_properties_setf(props, PW_KEY_NODE_RATE,
-					"1/%u", q.denom);
-			pw_properties_setf(props, PW_KEY_NODE_LATENCY,
-					"%u/%u", q.num, q.denom);
-		}
-	}
-	if ((str = getenv("PIPEWIRE_LATENCY")) != NULL)
-		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
-	if ((str = getenv("PIPEWIRE_RATE")) != NULL)
-		pw_properties_set(props, PW_KEY_NODE_RATE, str);
-
 	if (pw_properties_get(props, PW_KEY_STREAM_IS_LIVE) == NULL)
 		pw_properties_set(props, PW_KEY_STREAM_IS_LIVE, "true");
 	if (pw_properties_get(props, PW_KEY_NODE_NAME) == NULL && extra) {
@@ -1496,6 +1524,10 @@ stream_new(struct pw_context *context, const char *name,
 			str = name;
 		pw_properties_set(props, PW_KEY_NODE_NAME, str);
 	}
+	if ((pw_properties_get(props, PW_KEY_NODE_WANT_DRIVER) == NULL))
+		pw_properties_set(props, PW_KEY_NODE_WANT_DRIVER, "true");
+
+	pw_context_conf_update_props(context, "stream.properties", props);
 
 	this->name = name ? strdup(name) : NULL;
 	this->node_id = SPA_ID_INVALID;
@@ -1513,9 +1545,6 @@ stream_new(struct pw_context *context, const char *name,
 	impl->allow_mlock = context->settings.mem_allow_mlock;
 	impl->warn_mlock = context->settings.mem_warn_mlock;
 
-	spa_hook_list_append(&impl->context->driver_listener_list,
-			&impl->context_listener,
-			&context_events, impl);
 	return impl;
 
 error_properties:
@@ -1566,7 +1595,7 @@ pw_stream_new_simple(struct pw_loop *loop,
 	if (props == NULL)
 		return NULL;
 
-	context = pw_context_new(loop, NULL, 0);
+	context = pw_context_new(loop, pw_properties_copy(props), 0);
 	if (context == NULL) {
 		res = -errno;
 		goto error_cleanup;
@@ -1622,16 +1651,16 @@ static int stream_disconnect(struct stream *impl)
 
 	impl->disconnecting = true;
 
-	if (impl->node)
-		pw_impl_node_set_active(impl->node, false);
+	if (stream->node)
+		pw_impl_node_set_active(stream->node, false);
 
 	if (stream->proxy) {
 		pw_proxy_destroy(stream->proxy);
 		stream->proxy = NULL;
 	}
 
-	if (impl->node)
-		pw_impl_node_destroy(impl->node);
+	if (stream->node)
+		pw_impl_node_destroy(stream->node);
 
 	if (impl->disconnect_core) {
 		impl->disconnect_core = false;
@@ -1649,7 +1678,7 @@ void pw_stream_destroy(struct pw_stream *stream)
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	struct control *c;
 
-	ensure_loop(impl->context->main_loop, return);
+	ensure_loop(impl->main_loop, return);
 
 	pw_log_debug("%p: destroy", stream);
 
@@ -1681,8 +1710,6 @@ void pw_stream_destroy(struct pw_stream *stream)
 	spa_hook_list_clean(&impl->hooks);
 	spa_hook_list_clean(&stream->listener_list);
 
-	spa_hook_remove(&impl->context_listener);
-
 	if (impl->data.context)
 		pw_context_destroy(impl->data.context);
 
@@ -1690,10 +1717,22 @@ void pw_stream_destroy(struct pw_stream *stream)
 	free(impl);
 }
 
+static int
+do_remove_callbacks(struct spa_loop *loop,
+                 bool async, uint32_t seq, const void *data, size_t size, void *user_data)
+{
+	struct stream *impl = user_data;
+	spa_zero(impl->rt_callbacks);
+	return 0;
+}
+
 static void hook_removed(struct spa_hook *hook)
 {
 	struct stream *impl = hook->priv;
-	spa_zero(impl->rt_callbacks);
+	if (impl->data_loop)
+		pw_loop_invoke(impl->data_loop, do_remove_callbacks, 1, NULL, 0, true, impl);
+	else
+		spa_zero(impl->rt_callbacks);
 	hook->priv = NULL;
 	hook->removed = NULL;
 }
@@ -1706,7 +1745,7 @@ void pw_stream_add_listener(struct pw_stream *stream,
 {
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 
-	ensure_loop(impl->context->main_loop);
+	ensure_loop(impl->main_loop);
 
 	spa_hook_list_append(&stream->listener_list, listener, events, data);
 
@@ -1744,7 +1783,7 @@ int pw_stream_update_properties(struct pw_stream *stream, const struct spa_dict
 	int changed, res = 0;
 	struct match match;
 
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 
 	changed = pw_properties_update(stream->properties, dict);
 	if (!changed)
@@ -1754,8 +1793,8 @@ int pw_stream_update_properties(struct pw_stream *stream, const struct spa_dict
 	pw_context_conf_section_match_rules(impl->context, "stream.rules",
 			&stream->properties->dict, execute_match, &match);
 
-	if (impl->node)
-		res = pw_impl_node_update_properties(impl->node,
+	if (stream->node)
+		res = pw_impl_node_update_properties(stream->node,
 				match.count == 0 ?
 					dict :
 					&stream->properties->dict);
@@ -1852,14 +1891,15 @@ pw_stream_connect(struct pw_stream *stream,
 	struct pw_impl_factory *factory;
 	struct pw_properties *props = NULL;
 	const char *str;
+	struct match match;
 	uint32_t i;
 	int res;
 
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 
 	pw_log_debug("%p: connect target:%d", stream, target_id);
 
-	if (impl->node != NULL || stream->state != PW_STREAM_STATE_UNCONNECTED)
+	if (stream->node != NULL || stream->state != PW_STREAM_STATE_UNCONNECTED)
 		return -EBUSY;
 
 	impl->direction =
@@ -1873,6 +1913,7 @@ pw_stream_connect(struct pw_stream *stream,
 		impl->node_methods.process = impl_node_process_output;
 
 	impl->process_rt = SPA_FLAG_IS_SET(flags, PW_STREAM_FLAG_RT_PROCESS);
+	impl->early_process = SPA_FLAG_IS_SET(flags, PW_STREAM_FLAG_EARLY_PROCESS);
 
 	impl->impl_node.iface = SPA_INTERFACE_INIT(
 			SPA_TYPE_INTERFACE_Node,
@@ -1897,7 +1938,7 @@ pw_stream_connect(struct pw_stream *stream,
 	impl->info.flags = SPA_NODE_FLAG_RT;
 	/* if the callback was not marked RT_PROCESS, we will offload
 	 * the process callback in the main thread and we are ASYNC */
-	if (!impl->process_rt)
+	if (!impl->process_rt || SPA_FLAG_IS_SET(flags, PW_STREAM_FLAG_ASYNC))
 		impl->info.flags |= SPA_NODE_FLAG_ASYNC;
 	impl->info.props = &stream->properties->dict;
 	impl->params[NODE_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, 0);
@@ -1924,6 +1965,7 @@ pw_stream_connect(struct pw_stream *stream,
 	impl->port_params[PORT_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
 	impl->port_params[PORT_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
 	impl->port_params[PORT_Latency] = SPA_PARAM_INFO(SPA_PARAM_Latency, SPA_PARAM_INFO_WRITE);
+	impl->port_params[PORT_Tag] = SPA_PARAM_INFO(SPA_PARAM_Tag, SPA_PARAM_INFO_WRITE);
 	impl->port_info.props = &impl->port_props->dict;
 	impl->port_info.params = impl->port_params;
 	impl->port_info.n_params = N_PORT_PARAMS;
@@ -1943,49 +1985,71 @@ pw_stream_connect(struct pw_stream *stream,
 	impl->driving = false;
 	impl->trigger = false;
 	impl->using_trigger = false;
-	stream_set_state(stream, PW_STREAM_STATE_CONNECTING, NULL);
+	stream_set_state(stream, PW_STREAM_STATE_CONNECTING, 0, NULL);
 
-	if ((str = getenv("PIPEWIRE_NODE")) != NULL)
-		pw_properties_set(stream->properties, PW_KEY_TARGET_OBJECT, str);
-	else if (target_id != PW_ID_ANY)
+	if (target_id != PW_ID_ANY)
 		/* XXX this is deprecated but still used by the portal and its apps */
-		pw_properties_setf(stream->properties, PW_KEY_NODE_TARGET, "%d", target_id);
+		if (pw_properties_get(stream->properties, PW_KEY_NODE_TARGET) == NULL)
+			pw_properties_setf(stream->properties, PW_KEY_NODE_TARGET, "%d", target_id);
+	if (flags & PW_STREAM_FLAG_AUTOCONNECT)
+		if (pw_properties_get(stream->properties, PW_KEY_NODE_AUTOCONNECT) == NULL)
+			pw_properties_set(stream->properties, PW_KEY_NODE_AUTOCONNECT, "true");
+	if (flags & PW_STREAM_FLAG_EXCLUSIVE)
+		if (pw_properties_get(stream->properties, PW_KEY_NODE_EXCLUSIVE) == NULL)
+			pw_properties_set(stream->properties, PW_KEY_NODE_EXCLUSIVE, "true");
+	if (flags & PW_STREAM_FLAG_DONT_RECONNECT)
+		if (pw_properties_get(stream->properties, PW_KEY_NODE_DONT_RECONNECT) == NULL)
+			pw_properties_set(stream->properties, PW_KEY_NODE_DONT_RECONNECT, "true");
 
-	if ((flags & PW_STREAM_FLAG_AUTOCONNECT) &&
-	    pw_properties_get(stream->properties, PW_KEY_NODE_AUTOCONNECT) == NULL) {
-		str = getenv("PIPEWIRE_AUTOCONNECT");
-		pw_properties_set(stream->properties, PW_KEY_NODE_AUTOCONNECT, str ? str : "true");
-	}
 	if (flags & PW_STREAM_FLAG_DRIVER)
 		pw_properties_set(stream->properties, PW_KEY_NODE_DRIVER, "true");
-	if ((pw_properties_get(stream->properties, PW_KEY_NODE_WANT_DRIVER) == NULL))
-		pw_properties_set(stream->properties, PW_KEY_NODE_WANT_DRIVER, "true");
-
-	if (flags & PW_STREAM_FLAG_EXCLUSIVE)
-		pw_properties_set(stream->properties, PW_KEY_NODE_EXCLUSIVE, "true");
-	if (flags & PW_STREAM_FLAG_DONT_RECONNECT)
-		pw_properties_set(stream->properties, PW_KEY_NODE_DONT_RECONNECT, "true");
 	if (flags & PW_STREAM_FLAG_TRIGGER) {
 		pw_properties_set(stream->properties, PW_KEY_NODE_TRIGGER, "true");
 		impl->trigger = true;
 	}
-
-	if ((str = pw_properties_get(stream->properties, "mem.warn-mlock")) != NULL)
-		impl->warn_mlock = pw_properties_parse_bool(str);
-
 	if ((pw_properties_get(stream->properties, PW_KEY_MEDIA_CLASS) == NULL)) {
 		const char *media_type = pw_properties_get(stream->properties, PW_KEY_MEDIA_TYPE);
 		pw_properties_setf(stream->properties, PW_KEY_MEDIA_CLASS, "Stream/%s/%s",
 				direction == PW_DIRECTION_INPUT ? "Input" : "Output",
 				media_type ? media_type : get_media_class(impl));
 	}
-
 	if ((str = pw_properties_get(stream->properties, PW_KEY_FORMAT_DSP)) != NULL)
 		pw_properties_set(impl->port_props, PW_KEY_FORMAT_DSP, str);
 	else if (impl->media_type == SPA_MEDIA_TYPE_application &&
 	    impl->media_subtype == SPA_MEDIA_SUBTYPE_control)
 		pw_properties_set(impl->port_props, PW_KEY_FORMAT_DSP, "8 bit raw midi");
 
+	match = MATCH_INIT(stream);
+	pw_context_conf_section_match_rules(impl->context, "stream.rules",
+			&stream->properties->dict, execute_match, &match);
+
+	if ((str = getenv("PIPEWIRE_NODE")) != NULL)
+		pw_properties_set(stream->properties, PW_KEY_TARGET_OBJECT, str);
+	if ((str = getenv("PIPEWIRE_AUTOCONNECT")) != NULL)
+		pw_properties_set(stream->properties,
+				PW_KEY_NODE_AUTOCONNECT, spa_atob(str) ? "true" : "false");
+
+	if ((str = getenv("PIPEWIRE_PROPS")) != NULL)
+		pw_properties_update_string(stream->properties, str, strlen(str));
+	if ((str = getenv("PIPEWIRE_QUANTUM")) != NULL) {
+		struct spa_fraction q;
+		if (sscanf(str, "%u/%u", &q.num, &q.denom) == 2 && q.denom != 0) {
+			pw_properties_setf(stream->properties, PW_KEY_NODE_FORCE_RATE,
+					"1/%u", q.denom);
+			pw_properties_setf(stream->properties, PW_KEY_NODE_FORCE_QUANTUM,
+					"%u", q.num);
+		}
+	}
+	if ((str = getenv("PIPEWIRE_LATENCY")) != NULL)
+		pw_properties_set(stream->properties, PW_KEY_NODE_LATENCY, str);
+	if ((str = getenv("PIPEWIRE_RATE")) != NULL)
+		pw_properties_set(stream->properties, PW_KEY_NODE_RATE, str);
+
+	if ((str = pw_properties_get(stream->properties, "mem.warn-mlock")) != NULL)
+		impl->warn_mlock = pw_properties_parse_bool(str);
+	if ((str = pw_properties_get(stream->properties, "mem.allow-mlock")) != NULL)
+		impl->allow_mlock = pw_properties_parse_bool(str);
+
 	impl->port_info.props = &impl->port_props->dict;
 
 	if (stream->core == NULL) {
@@ -2012,6 +2076,7 @@ pw_stream_connect(struct pw_stream *stream,
 	    pw_properties_parse_bool(str)) {
 		pw_properties_set(props, "resample.peaks", "true");
 		pw_properties_set(props, "channelmix.normalize", "true");
+		pw_properties_set(props, PW_KEY_PORT_IGNORE_LATENCY, "true");
 	}
 
 	if (impl->media_type == SPA_MEDIA_TYPE_audio) {
@@ -2024,32 +2089,34 @@ pw_stream_connect(struct pw_stream *stream,
 		pw_properties_setf(props, "adapt.follower.spa-node", "pointer:%p",
 				&impl->impl_node);
 		pw_properties_set(props, "object.register", "false");
-		impl->node = pw_impl_factory_create_object(factory,
+		stream->node = pw_impl_factory_create_object(factory,
 				NULL,
 				PW_TYPE_INTERFACE_Node,
 				PW_VERSION_NODE,
 				props,
 				0);
 		props = NULL;
-		if (impl->node == NULL) {
+		if (stream->node == NULL) {
 			res = -errno;
 			goto error_node;
 		}
 	} else {
-		impl->node = pw_context_create_node(impl->context, props, 0);
+		stream->node = pw_context_create_node(impl->context, props, 0);
 		props = NULL;
-		if (impl->node == NULL) {
+		if (stream->node == NULL) {
 			res = -errno;
 			goto error_node;
 		}
-		pw_impl_node_set_implementation(impl->node, &impl->impl_node);
+		pw_impl_node_set_implementation(stream->node, &impl->impl_node);
 	}
-	pw_impl_node_set_active(impl->node,
+	pw_impl_node_set_active(stream->node,
 			!SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_INACTIVE));
 
-	pw_log_debug("%p: export node %p", stream, impl->node);
+	impl->data_loop = stream->node->data_loop;
+
+	pw_log_debug("%p: export node %p", stream, stream->node);
 	stream->proxy = pw_core_export(stream->core,
-			PW_TYPE_INTERFACE_Node, NULL, impl->node, 0);
+			PW_TYPE_INTERFACE_Node, NULL, stream->node, 0);
 	if (stream->proxy == NULL) {
 		res = -errno;
 		goto error_proxy;
@@ -2057,7 +2124,9 @@ pw_stream_connect(struct pw_stream *stream,
 
 	pw_proxy_add_listener(stream->proxy, &stream->proxy_listener, &proxy_events, stream);
 
-	pw_impl_node_add_listener(impl->node, &stream->node_listener, &node_events, stream);
+	pw_impl_node_add_listener(stream->node, &stream->node_listener, &node_events, stream);
+	pw_impl_node_add_rt_listener(stream->node, &stream->node_rt_listener,
+			&node_rt_events, stream);
 
 	return 0;
 
@@ -2086,7 +2155,7 @@ SPA_EXPORT
 int pw_stream_disconnect(struct pw_stream *stream)
 {
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 	return stream_disconnect(impl);
 }
 
@@ -2096,11 +2165,11 @@ int pw_stream_set_error(struct pw_stream *stream,
 {
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 
 	if (res < 0) {
+		spa_autofree char *value = NULL;
 		va_list args;
-		char *value;
 		int r;
 
 		va_start(args, error);
@@ -2111,9 +2180,7 @@ int pw_stream_set_error(struct pw_stream *stream,
 
 		if (stream->proxy)
 			pw_proxy_error(stream->proxy, res, value);
-		stream_set_state(stream, PW_STREAM_STATE_ERROR, value);
-
-		free(value);
+		stream_set_state(stream, PW_STREAM_STATE_ERROR, res, value);
 	}
 	return res;
 }
@@ -2126,7 +2193,7 @@ int pw_stream_update_params(struct pw_stream *stream,
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	int res;
 
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 
 	pw_log_debug("%p: update params", stream);
 	if ((res = update_params(impl, SPA_ID_INVALID, params, n_params)) < 0)
@@ -2142,7 +2209,7 @@ static inline int stream_set_param(struct stream *impl, uint32_t id, const struc
 {
 	int res = 0;
 	impl->in_set_param++;
-	res = pw_impl_node_set_param(impl->node, id, 0, param);
+	res = pw_impl_node_set_param(impl->this.node, id, 0, param);
 	impl->in_set_param--;
 	return res;
 }
@@ -2151,9 +2218,9 @@ SPA_EXPORT
 int pw_stream_set_param(struct pw_stream *stream, uint32_t id, const struct spa_pod *param)
 {
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 
-	if (impl->node == NULL)
+	if (stream->node == NULL)
 		return -EIO;
 
 	return stream_set_param(impl, id, param);
@@ -2170,9 +2237,9 @@ int pw_stream_set_control(struct pw_stream *stream, uint32_t id, uint32_t n_valu
 	struct spa_pod *pod;
 	struct control *c;
 
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 
-	if (impl->node == NULL)
+	if (stream->node == NULL)
 		return -EIO;
 
 	va_start(varargs, values);
@@ -2239,14 +2306,14 @@ int pw_stream_set_active(struct pw_stream *stream, bool active)
 {
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 
-	ensure_loop(impl->context->main_loop, return -EIO);
+	ensure_loop(impl->main_loop, return -EIO);
 
 	pw_log_debug("%p: active:%d", stream, active);
 
-	if (impl->node == NULL)
+	if (stream->node == NULL)
 		return -EIO;
 
-	pw_impl_node_set_active(impl->node, active);
+	pw_impl_node_set_active(stream->node, active);
 
 	if (!active || impl->drained)
 		impl->drained = impl->draining = false;
@@ -2273,14 +2340,15 @@ int pw_stream_get_time_n(struct pw_stream *stream, struct pw_time *time, size_t
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	uintptr_t seq1, seq2;
 	uint32_t buffered, quantum, index;
+	int32_t avail_buffers;
 
 	do {
-		seq1 = SEQ_READ(impl->seq);
+		seq1 = SPA_SEQ_READ(impl->seq);
 		memcpy(time, &impl->time, SPA_MIN(size, sizeof(struct pw_time)));
 		buffered = impl->rate_queued;
 		quantum = impl->quantum;
-		seq2 = SEQ_READ(impl->seq);
-	} while (!SEQ_READ_SUCCESS(seq1, seq2));
+		seq2 = SPA_SEQ_READ(impl->seq);
+	} while (!SPA_SEQ_READ_SUCCESS(seq1, seq2));
 
 	if (impl->direction == SPA_DIRECTION_INPUT)
 		time->queued = (int64_t)(time->queued - impl->dequeued.outcount);
@@ -2291,19 +2359,23 @@ int pw_stream_get_time_n(struct pw_stream *stream, struct pw_time *time, size_t
 	time->delay += (impl->latency.min_rate + impl->latency.max_rate) / 2;
 	time->delay += ((impl->latency.min_ns + impl->latency.max_ns) / 2) * time->rate.denom / SPA_NSEC_PER_SEC;
 
+	avail_buffers = spa_ringbuffer_get_read_index(&impl->dequeued.ring, &index);
+	avail_buffers = SPA_CLAMP(avail_buffers, 0, (int32_t)impl->n_buffers);
+
 	if (size >= offsetof(struct pw_time, queued_buffers))
 		time->buffered = buffered;
 	if (size >= offsetof(struct pw_time, avail_buffers))
-		time->queued_buffers = spa_ringbuffer_get_read_index(&impl->queued.ring, &index);
+		time->queued_buffers = impl->n_buffers - avail_buffers;
 	if (size >= sizeof(struct pw_time))
-		time->avail_buffers = spa_ringbuffer_get_read_index(&impl->dequeued.ring, &index);
+		time->avail_buffers = avail_buffers;
 
 	pw_log_trace_fp("%p: %"PRIi64" %"PRIi64" %"PRIu64" %d/%d %"PRIu64" %"
-			PRIu64" %"PRIu64" %"PRIu64" %"PRIu64, stream,
+			PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %d/%d", stream,
 			time->now, time->delay, time->ticks,
 			time->rate.num, time->rate.denom, time->queued,
 			impl->dequeued.outcount, impl->dequeued.incount,
-			impl->queued.outcount, impl->queued.incount);
+			impl->queued.outcount, impl->queued.incount,
+			avail_buffers, impl->n_buffers);
 	return 0;
 }
 
@@ -2329,11 +2401,12 @@ struct pw_buffer *pw_stream_dequeue_buffer(struct pw_stream *stream)
 		errno = -res;
 		return NULL;
 	}
-	pw_log_trace_fp("%p: dequeue buffer %d size:%"PRIu64, stream, b->id, b->this.size);
+	pw_log_trace_fp("%p: dequeue buffer %d size:%"PRIu64" req:%"PRIu64,
+			stream, b->id, b->this.size, b->this.requested);
 
 	if (b->busy && impl->direction == SPA_DIRECTION_OUTPUT) {
-		if (ATOMIC_INC(b->busy->count) > 1) {
-			ATOMIC_DEC(b->busy->count);
+		if (SPA_ATOMIC_INC(b->busy->count) > 1) {
+			SPA_ATOMIC_DEC(b->busy->count);
 			queue_push(impl, &impl->dequeued, b);
 			pw_log_trace_fp("%p: buffer busy", stream);
 			errno = EBUSY;
@@ -2351,16 +2424,17 @@ int pw_stream_queue_buffer(struct pw_stream *stream, struct pw_buffer *buffer)
 	int res;
 
 	if (b->busy)
-		ATOMIC_DEC(b->busy->count);
+		SPA_ATOMIC_DEC(b->busy->count);
 
-	pw_log_trace_fp("%p: queue buffer %d", stream, b->id);
+	pw_log_trace_fp("%p: queue buffer %d size:%"PRIu64, stream, b->id,
+			b->this.size);
 	if ((res = queue_push(impl, &impl->queued, b)) < 0)
 		return res;
 
 	if (impl->direction == SPA_DIRECTION_OUTPUT &&
 	    impl->driving && !impl->using_trigger) {
 		pw_log_debug("deprecated: use pw_stream_trigger_process() to drive the stream.");
-		res = pw_loop_invoke(impl->context->data_loop,
+		res = pw_loop_invoke(impl->data_loop,
 			do_trigger_deprecated, 1, NULL, 0, false, impl);
 	}
 	return res;
@@ -2372,12 +2446,21 @@ do_flush(struct spa_loop *loop,
 {
 	struct stream *impl = user_data;
 	struct buffer *b;
+	struct queue *from, *to;
 
 	pw_log_trace_fp("%p: flush", impl);
+
+	if (impl->direction == SPA_DIRECTION_OUTPUT) {
+		from = &impl->queued;
+		to = &impl->dequeued;
+	} else {
+		from = &impl->dequeued;
+		to = &impl->queued;
+	}
 	do {
-		b = queue_pop(impl, &impl->queued);
+		b = queue_pop(impl, from);
 		if (b != NULL)
-			queue_push(impl, &impl->dequeued, b);
+			queue_push(impl, to, b);
 	}
 	while (b);
 
@@ -2402,14 +2485,14 @@ int pw_stream_flush(struct pw_stream *stream, bool drain)
 {
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 
-	if (impl->node == NULL)
+	if (stream->node == NULL)
 		return -EIO;
 
-	pw_loop_invoke(impl->context->data_loop,
+	pw_loop_invoke(impl->data_loop,
 			drain ? do_drain : do_flush, 1, NULL, 0, true, impl);
 
 	if (!drain)
-		spa_node_send_command(impl->node->node,
+		spa_node_send_command(stream->node->node,
 				&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Flush));
 	return 0;
 }
@@ -2422,7 +2505,7 @@ bool pw_stream_is_driving(struct pw_stream *stream)
 }
 
 static int
-do_trigger_process(struct spa_loop *loop,
+do_trigger_driver(struct spa_loop *loop,
                  bool async, uint32_t seq, const void *data, size_t size, void *user_data)
 {
 	struct stream *impl = user_data;
@@ -2457,20 +2540,21 @@ int pw_stream_trigger_process(struct pw_stream *stream)
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	int res = 0;
 
-	pw_log_trace_fp("%p", impl);
+	pw_log_trace_fp("%p: trigger:%d driving:%d", impl, impl->trigger, impl->driving);
 
 	/* flag to check for old or new behaviour */
 	impl->using_trigger = true;
 
-	if (!impl->driving && !impl->trigger) {
-		res = pw_loop_invoke(impl->context->main_loop,
-			do_trigger_request_process, 1, NULL, 0, false, impl);
-	} else {
+	if (impl->trigger) {
+		pw_impl_node_trigger(stream->node);
+	} else if (impl->driving) {
 		if (!impl->process_rt)
 			call_process(impl);
-
-		res = pw_loop_invoke(impl->context->data_loop,
-			do_trigger_process, 1, NULL, 0, false, impl);
+		res = pw_loop_invoke(impl->data_loop,
+			do_trigger_driver, 1, NULL, 0, false, impl);
+	} else {
+		res = pw_loop_invoke(impl->main_loop,
+			do_trigger_request_process, 1, NULL, 0, false, impl);
 	}
 	return res;
 }
